(* Ocamllex scanner for MicroC *)

(* This code was original from MicroC and modified by:
Keir Lauritzen, Leon Song, Darren Hakimi, and Guy Yardeni*)

{ 
	open Parser 
	type is_pat = REGEX | NORMAL
	let state_ref = ref NORMAL

let unescape s =
    	Scanf.sscanf ("\"" ^ s ^ "\"") "%S%!" (fun x -> x)
}

let alphabet = ['a'-'z' 'A'-'Z']
let escape = '\\' ['\\' ''' '"' 'n' 'r' 't']
let escape_char = ''' (escape) '''
let ascii = ([' '-'!' '#'-'[' ']'-'~'])
let digit = ['0'-'9']
let string_lit = '"' ( (ascii | escape)* as s) '"'
let character = ''' ( ascii | digit ) '''


(* define some common character classes here *)
let digits = ['0' - '9']+
let id = ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']*
(* Identifiers can start with alphabetic characters only *)

rule token pat = parse
  [' ' '\t' '\r' '\n'] { token pat lexbuf } (* Whitespace *)
| "/*"     { comment pat lexbuf }           (* Comments *)
| '('      { LPAREN }
| ')'      { RPAREN }
| '{'      { LBRACE }
| '}'      { RBRACE }
| ';'      { SEMI }
| ','      { COMMA }
| '+'      { PLUS }
| '-'      { MINUS }
| '['	   { LBRACK }
| ']'	   { RBRACK }
| '*'      { TIMES }
| '/'      { DIVIDE }
| '='      { ASSIGN }
| "=="     { EQ }
| "!="     { NEQ }
| '<'      { LT }
| "<="     { LEQ }
| ">"      { GT }
| ">="     { GEQ }
| "&&"     { AND }
| "||"     { OR }
| "!"      { NOT }
| "if"     { IF }
| "else"   { ELSE }
| "for"    { FOR }
| "while"  { WHILE }
| "return" { RETURN }
| "function" { FUNCTION }
| "//"       { pat := REGEX ; FSLASH }
| "BEGIN"   { BEGIN }
| "END"     { END }
| "bool"    { BOOL }
| "TRUE"    { TRUE }
| "FALSE"   { FALSE }
| "void"   { VOID }
| "array_float"  { ARRAY_F }
| "array_string" { ARRAY_S }
| "float"  { FLOAT }
| "string" { STRING }
| "RS"     { RS }
| "FS"     { FS }
| "NPAT"     { NPAT }
| '$''0'     { RECORD }
| '$'['1' - '9']['0' - '9']+  { RFIELD }
| "__patterns"  { PATTERNS }
| "__action"['0'-'9']+   { ACTIONS }
| "__actions"   { ACTIONS2 }
| "destroy()"   { DESTROY }
| "__init()" { INIT}
| ['-']?digits as lxm {FLOAT_L(float_of_string lxm)}
| ['-']?['0' - '9']+['.']['0' - '9']+ as lxm {FLOAT_L(float_of_string lxm)}
| ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']* as lxm { ID(lxm) }
| '"' [^ '"']* '"' as lxm { STRING_L(lxm) }
| eof { EOF }
| _ as char { raise (Failure("illegal character " ^ Char.escaped char)) }

and comment pat = parse
  "*/" { token pat lexbuf }
| _    { comment pat lexbuf }

and regex pat = parse
| "//" { pat := NORMAL ; FSLASH }
| ['\\']['"' '.' '?' '|' '^' ']' '[' '(' ')' '-' '\\' '$' '*'] as lxm{ REGEX_STRING(lxm) }
| '.' { PERIOD }
| '?' { QUEST }
| '^' { CARROT }
| '|' { VERT }
| '-' { RMINUS }
| '*' { TIMES }
| '[' { LBRACK }
| ']' { RBRACK }
| '(' { LPAREN }
| ')' { RPAREN }
| [^'"' '/' '.' '?' '|' '^' ']' '[' '(' ')' '-' '\\' '$' '*'] as lxm { REGEX_STRING((Char.escaped lxm)) }


(*The function to get the next token; checks to see if it is scanning a regex pattern block*)
{
	let next_token lexbuf = match !state_ref with
	    | NORMAL -> token state_ref lexbuf
	    | REGEX -> regex state_ref lexbuf
}



